#include "header.h"



// PRJ file's TERR chunk to uncompressed 16-bit grayscale PNG files 
#pragma intrinsic( _byteswap_ulong, _byteswap_ushort, memcpy )
void ExportPNG( __in char* png_out_name, __in HEIGHTMAP_INFO* i, __in DWORD map_id )
{
	BYTE* png_buf; 
	DWORD png_size;
	BYTE* filtered_data;
	DWORD filtered_data_size;
	DWORD deflate_data_size;
	DWORD adler;

	// calc sizes
	// ( width * height * 2 byte-per-pixel ) + 1 filter byte per scanline
	filtered_data_size = i->header->width * sizeof(WORD) * i->header->height + i->header->height;
	// (deflate_header+footer) + filtered_size + ( num_chunks * chunk_header_size )
	deflate_data_size = 6 + filtered_data_size + ((( filtered_data_size + 0x0000FFFE ) / 0x0000FFFF ) * 5 ); 
	// png = SIG + IHDR + IDAT + IEND
	png_size = (8) + (13 + 12) + (deflate_data_size + 12) + (0 + 12); 
	
	// alloc
	png_buf = (BYTE*) VirtualAlloc( NULL, png_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE );
	filtered_data = (BYTE*) VirtualAlloc( NULL, filtered_data_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE );
	if( ( png_buf == NULL ) || ( filtered_data == NULL ) ){ 
		Printf( "Error: Out of memory while processing %s.\r\n", png_out_name );
		if( png_buf != NULL ) VirtualFree( png_buf, 0, MEM_RELEASE );
		if( filtered_data != NULL ) VirtualFree( filtered_data, 0, MEM_RELEASE );
		return;
	}

	// get filtered_data from heightmap structures
	BYTE* p = (BYTE*)filtered_data;
	for( DWORD y = 0; y < i->header->height; y++ ){
		*p++ = 0x00; // PNG filter byte ( no filter )
		for( DWORD x = 0; x < i->header->width; x++ ){
			LAYER1::BLOCK* pBlock1 = &i->layer1->blocks[ ( ( ( y >> 3 ) * ( ( i->header->width + 7 ) >> 3 ) ) + ( x >> 3 ) )  + ( i->header->num_blocks_layer1 * map_id ) ];  
			LAYER2::BLOCK* pBlock2 = &i->layer2->blocks[ pBlock1->layer2_offset >> 6 ];
			*((WORD*)p) = _byteswap_ushort( ((WORD)( pBlock1->base_height + ( ( pBlock2->delta[ ( y & 7 ) ][ ( x & 7 ) ] ) << 7 ) )) + i->sea_level ); 
			p+=2; 
		}
	}
	adler = adler32( filtered_data, filtered_data_size );

	// build uncompressed 16-bit grayscale PNG
	*((DWORD*)( &png_buf[0]  )) = 0x474E5089; // png_magic_1
	*((DWORD*)( &png_buf[4]  )) = 0x0A1A0A0D; // png_magic_2
	
	*((DWORD*)( &png_buf[8]  )) = 0x0D000000; // IHDR_len
	*((DWORD*)( &png_buf[12] )) = 0x52444849; // "IHDR"
	*((DWORD*)( &png_buf[16] )) = _byteswap_ulong( i->header->width ); // width
	*((DWORD*)( &png_buf[20] )) = _byteswap_ulong( i->header->height ); // height
	*((BYTE* )( &png_buf[24] )) = 0x10; // bit-depth
	*((DWORD*)( &png_buf[25] )) = 0x00000000; // color, compression, filter, and interlace
	*((DWORD*)( &png_buf[29] )) = _byteswap_ulong( crc32( &png_buf[12], 17 ) ); // checksum of IHDR chunk

	*((DWORD*)( &png_buf[33] )) = _byteswap_ulong( deflate_data_size ); // IDAT_len
	*((DWORD*)( &png_buf[37] )) = 0x54414449; // "IDAT"
	*((BYTE* )( &png_buf[41] )) = 0x78; // Compression Method and flags ( 0x78 = deflate with a 32 KiB window )
	*((BYTE* )( &png_buf[42] )) = 0x01; // Additional flags | check bits ( 0x01 )

	DWORD dst_off;
	DWORD src_off;
	// do all but the last chunk
	for( dst_off = 43, src_off = 0; (src_off + 0xFFFF ) < filtered_data_size; src_off += 0x0000FFFF, dst_off += ( 0x0000FFFF + 5 ) ) 
	{
		*((BYTE* )( &png_buf[ dst_off + 0 ] )) = 0x00; // BFINAL | BTYPE
		*((DWORD*)( &png_buf[ dst_off + 1 ] )) = 0x0000FFFF; // chunk_len | not chunk_len
		memcpy( &png_buf[ dst_off + 5 ], &filtered_data[ src_off ], 0x0000FFFF );
	}

	// do last deflate chunk
	WORD chunk_len = (WORD)( filtered_data_size - src_off );
	*((BYTE* )( &png_buf[ dst_off + 0 ] )) = 0x01; // BFINAL | BTYPE
	*((WORD* )( &png_buf[ dst_off + 1 ] )) = chunk_len; // chunk_len
	*((WORD* )( &png_buf[ dst_off + 3 ] )) = ~chunk_len; // one's complement of chunk_len
	memcpy( &png_buf[ dst_off + 5 ], &filtered_data[ src_off ], chunk_len );
	dst_off += ( chunk_len + 5 );
	*((DWORD*)( &png_buf[ dst_off + 0 ]  )) = _byteswap_ulong( adler ); // checksum deflate-compressed datastream
	*((DWORD*)( &png_buf[ dst_off + 4 ]  )) = _byteswap_ulong( crc32( &png_buf[37], deflate_data_size + 4 ) ); // crc of IDAT chunk

	*((DWORD*)( &png_buf[ dst_off + 8 ]  )) = 0x02000000;
	*((DWORD*)( &png_buf[ dst_off + 12 ] )) = 0x52526574; // "teRR" 
	*((WORD* )( &png_buf[ dst_off + 16 ] )) = i->sea_level;
	*((DWORD*)( &png_buf[ dst_off + 18 ] )) = _byteswap_ulong( crc32( &png_buf[ dst_off + 12 ], 6 ) );

	*((DWORD*)( &png_buf[ dst_off + 22 ]  )) = 0x00000000; // IEND_len
	*((DWORD*)( &png_buf[ dst_off + 26 ] )) = 0x444E4549; // "IEND"
	*((DWORD*)( &png_buf[ dst_off + 30 ] )) = 0x826042AE; // IEND_crc

	dst_off += 34;

	// write PNG
	if( SaveFile( png_out_name, png_buf, dst_off ) == FALSE )
	{
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", png_out_name );
		PrintSystemError( dwErr );
	}

	// free
	VirtualFree( png_buf, 0, MEM_RELEASE );
	VirtualFree( filtered_data, 0, MEM_RELEASE );
}


