#include "header.h"


// helper - check if pointer + offset is inside buf
#define check_seek(x) \
	if( ( p < i->file_buf ) || \
	( p >= (((BYTE*)i->file_buf) + i->file_size) ) || \
	( ( p + x ) >= (((BYTE*)i->file_buf) + i->file_size) ) || \
	( ( p + x ) < (BYTE*)i->file_buf ) ) { \
		pszError = "Unexpected end of file."; \
		goto lbl_error; \
	}

#pragma intrinsic( memcmp, abs )
BOOL GetTERR( __in  char* file_name, __out HEIGHTMAP_INFO* i )
{
	char* pszError;
	BYTE* p;
	signed int min_elevation = 0;

	// read-in file
	i->file_buf = ReadInFile( file_name, &i->file_size );
	if( i->file_buf == NULL ){
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", file_name );
		PrintSystemError( dwErr );
		return FALSE;
	}
	p = (BYTE*) i->file_buf;
	
	// is valid PRJ file?
	check_seek( 32 );
	if( memcmp( i->file_buf, "Dark Omen Battle file 1.10      ", 32 ) ){
		pszError = "Invalid file signature.";
		goto lbl_error;
	}

	p += 32; // skip signature "Dark Omen Battle file 1.10      "
	p += 4; // skip 'BASE'
	check_seek( 4 );
	p += (*((DWORD*)p)) + 4; // skip BASE data
	p += 4; // skip 'WATR'
	check_seek( 4 );
	p += (*((DWORD*)p)) + 4; // skip WATR data
	p += 4; // skip 'FURN'
	check_seek( 8 );
	p += (*((DWORD*) p)) + (*((DWORD*)(p + 4))) * 4 + 4; // skip FURN data
	p += 4; // skip 'INST'
	check_seek( 4 );
	p += (*((DWORD*)p)) + 12; // skip INST data

	// should be TERR chunk
	check_seek( sizeof(TERR_HEADER) );
	i->header = ( TERR_HEADER* ) p;

	// verify signature
	if( i->header->id[0] != 'T' || i->header->id[1] != 'E' || i->header->id[2] != 'R' || i->header->id[3] != 'R' )
	{
		pszError = "Bad TERR - signature not found.";
		goto lbl_error;
	}

	// todo: check header->size

	// check width and height ranges
	if( ( i->header->width == 0 ) || ( i->header->height == 0 ) || 
		( i->header->width >= 0x00010000 ) || ( i->header->height >= 0x00010000 ) ) // arbitrary
	{
		pszError = "Bad TERR - dimensions.";
		goto lbl_error;
	}

	// check number of layer1 blocks against the specified width & height
	// note: 8x8 blocks so width & height are rounded up to a multiple of 8
	if( ( ( ( ( i->header->width + 7 ) & 0xFFFFFFF8 ) * ( ( i->header->height + 7 ) & 0xFFFFFFF8 ) ) / 64 ) != i->header->num_blocks_layer1 )
	{
		pszError = "Bad TERR - number of layer1 blocks.";
		goto lbl_error;	
	}

	// there must be at least one layer2 block ??
	if( i->header->num_blocks_layer2 == 0 )
	{
		pszError = "Bad TERR - missing layer2 blocks.";
		goto lbl_error;
	}

	// LAYER1 part 
	p += sizeof(TERR_HEADER);
	i->layer1 = (LAYER1*) p;

	// check size against header num_blocks_layer1
	check_seek( 4 );
	if( i->layer1->size != ( ( i->header->num_blocks_layer1 * sizeof( LAYER1::BLOCK ) ) * 2 ) )
	{
		pszError = "Bad TERR - layer1 size.";
		goto lbl_error;
	}
		
	// LAYER2 part
	p += 4 + i->layer1->size;
	i->layer2 = (LAYER2*) p;
		
	// check size against header num_blocks_layer2
	check_seek( 4 );
	if( i->layer2->size != ( i->header->num_blocks_layer2 * sizeof( LAYER2::BLOCK ) ) )
	{
		pszError = "Bad TERR - layer2 size.";
		goto lbl_error;
	}
	
	// make sure TERR chunk is followed by ATTR chunk
	p += 4 + i->layer2->size;
	check_seek( 4 );
	if( p[0] != 'A' || p[1] != 'T' || p[2] != 'T' || p[3] != 'R' )
	{
		pszError = "ATTR chunk not found.";
		goto lbl_error;
	}

	// save for future use
	i->chunk_size = (DWORD)( p - (BYTE*)i->header );


	// get min_elevation and check offsets
	min_elevation = 0;
	for( DWORD k = 0; k < i->header->num_blocks_layer1 * 2; k++ )
	{
		min_elevation = min( min_elevation, i->layer1->blocks[k].base_height );
		if( ( i->layer1->blocks[k].layer2_offset % 64 ) || 
			( ( i->layer1->blocks[k].layer2_offset >> 6 ) >= i->header->num_blocks_layer2 ) ) 
		{
			pszError = "Bad TERR - bad layer2 offset.";
			goto lbl_error;
		}
	}	
	// 
	i->sea_level = (WORD)( abs( min_elevation ) );

	return TRUE;

lbl_error:
	Printf( "%s > %s\r\n", file_name, pszError );
	if( i->file_buf != NULL )
	{
		VirtualFree( i->file_buf, 0, MEM_RELEASE );
		i->file_buf = NULL;
	}
	return FALSE;
}