/* terrain heightmap
* 
* heightmap data is stored in some form of Block Truncation Coding
*
* The grayscale image is divided into 8x8 sub-blocks
* each block is represented by two "layers"
* Layer_1 is a 32-bit value that is used as the base level for all 64 pixels in the block
* Layer_2 is 64 8-bit values, one for each of the pixel in the block 
* The Layer_1 value is added to each of the Layer_2 values to decompress all 64 pixels in the block
* 
* The Layer_2 entries can be re-used by different blocks to save space ( dictionary compression )
* 
* data layout:
*		TERR_HEADER
*		LAYER1
*		LAYER2
*/

/* 
* note: LAYER1 contains two heightmaps ( of the same dimensions )
*	one for line of sight and one for movement? http://forum.dark-omen.org/tools/general-modding-questions-t1210.0.html
*	however, TERR_HEADER's width, height, and num_blocks_layer1 members are all missing the second heightmap 
*   and the size member isn't quite right either...
*/

#include <windows.h>


#include <pshpack1.h>
struct TERR_HEADER
{
	char id[4]; // 'TERR'
	DWORD size; 
	DWORD width; 
	DWORD height;
	DWORD num_blocks_layer2;
	DWORD num_blocks_layer1;
};
struct LAYER1
{
	DWORD size;
	struct BLOCK
	{
		signed int base_height;
		DWORD layer2_offset; // bytes from &layer2.blocks[0]
	} blocks[1];
};
struct LAYER2
{
	DWORD size;
	struct BLOCK
	{
		BYTE delta[8][8];
	} blocks[1];
};
#include <poppack.h>

// internal helper
struct HEIGHTMAP_INFO 
{
	void* file_buf;
	DWORD file_size;
	TERR_HEADER* header;
	LAYER1* layer1;
	LAYER2* layer2;
	DWORD chunk_size; // size of entire TERR chunk
	WORD sea_level; // color value of elevation zero
};

void SetSeaLevel( char* png_name, WORD sea_level );
void ShowSeaLevel( char* png_name );
void* lodepng_malloc(size_t size);
void* lodepng_realloc(void* ptr, size_t new_size);
void lodepng_free(void* ptr);
BOOL Import(  char* prj_name_in, char* prj_name_out, char* png_name_0, char* png_name_1 );
void ExportPNG( char* png_out_name, HEIGHTMAP_INFO* i, DWORD map_id );
void PrintSystemError( DWORD dwErrorCode );
void* ReadInFile( CHAR* szFileName, DWORD* pdwBytesRead );
BOOL SaveFile( char* file_name, void* buf, DWORD size );
void __cdecl Printf( char* fmt, ... );
BOOL GetTERR( char* file_name, HEIGHTMAP_INFO* info );
void ConsoleEnter();
void ConsoleExit();
unsigned int crc32( const unsigned char* buf, unsigned int len );
unsigned int adler32( const unsigned char* buf, unsigned int len );
DWORD ReadCmdLineArg( char* dst, DWORD dst_size, char* src, __out_opt DWORD* src_bytes_read );
DWORD ReadOutInteger( char* src, __out_opt DWORD* src_bytes_read );