#include "header.h"
#include "lodepng.h"


BYTE* get_chunk( BYTE* png_buf, DWORD png_size, DWORD chunk_id )
{
	for( BYTE* p = png_buf + 8; p <= png_buf + png_size - 12; p += _byteswap_ulong( *((DWORD*)p) ) + 12 ){
		if( *((DWORD*)( &p[4] )) == chunk_id ) return p; // "teRR"
	}
	return NULL;
}


#pragma intrinsic( _byteswap_ulong )
WORD* DecodePNG( char* file_name, DWORD* width, DWORD* height, WORD* sea_level )
{
	BYTE* png_buf;

	DWORD png_size;
	BYTE* raw_buf;
	LodePNGState state;

	// read-in file
	png_buf = (BYTE*) ReadInFile( file_name, &png_size );
	if( png_buf == NULL ){
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", file_name );
		PrintSystemError( dwErr );
		return NULL;
	}

	// decode png
	lodepng_state_init(&state);
	state.decoder.color_convert = 0;
	state.info_raw.colortype = LCT_GREY;
	state.info_raw.bitdepth = 16;
	lodepng_decode( &raw_buf, (unsigned int*)width, (unsigned int*)height, &state, png_buf, png_size );
	if( state.error ){
		Printf("%s > decode error %u: %s\n", file_name, state.error, lodepng_error_text( state.error ) );
		lodepng_free( raw_buf );
		raw_buf = NULL;
	}
	else{ // only allow 16-bit grayscale
		if( ( state.info_png.color.colortype != 0 ) || ( state.info_png.color.bitdepth != 16 ) ){
			Printf( "%s > detected %d-bit colortype = %d\r\n", file_name, state.info_png.color.bitdepth, state.info_png.color.colortype );
			Printf( "%s > Error: PNG not 16-bit grayscale ( colortype 0 ).\r\n", file_name );
			lodepng_free( raw_buf );
			raw_buf = NULL;
		}
		else{
			BYTE* p = get_chunk( png_buf, png_size, 0x52526574 ); // "teRR"
			*sea_level = ( ( p == NULL ) ? 0 : *((WORD*)(&p[8])) ); 
		}
	}

	// no longer need original png_data
	VirtualFree( png_buf, 0, MEM_RELEASE );
	
	lodepng_state_cleanup(&state);
	return (WORD*)raw_buf;
}

#pragma intrinsic( _byteswap_ushort, memcpy, memcmp )
BOOL Import(  char* prj_name_in, char* prj_name_out, char* png_name_0, char* png_name_1 )
{
	WORD* image_0;
	DWORD width_0;
	DWORD height_0;
	WORD sea_level_0;
	WORD* image_1;
	DWORD width_1;
	DWORD height_1;
	WORD sea_level_1;
	HEIGHTMAP_INFO info;
	TERR_HEADER* header;
	LAYER1* layer1;
	DWORD layer1_size;
	LAYER2* layer2;
	DWORD max_layer2_size;
	BYTE* out_buf;
	DWORD max_out_buf_size;
	DWORD out_size;
	DWORD& width = width_0;
	DWORD& height = height_0;
	WORD sea_level;
	DWORD block_index;
	WORD* image;
	BOOL bResult = FALSE;

	// get images
	image_0 = DecodePNG( png_name_0, &width_0, &height_0, &sea_level_0 );
	image_1 = DecodePNG( png_name_1, &width_1, &height_1, &sea_level_1 );
	if( ( image_0 == NULL ) || ( image_1 == NULL ) ) goto lbl_end;
	if( ( width_0 != width_1 ) || ( height_0 != height_1 ) )
	{
		Printf( "Error: %s (%dx%d) must be the same dimensions as %s (%dx%d).\r\n", png_name_0, width_0, height_0, png_name_1, width_1, height_1 );
		goto lbl_end;
	}
	if( sea_level_0 != sea_level_1 )
	{	// just-in-case sea_level info was stripped from one file but not the other?
		Printf( "Warning: %s sea-level (%04X) differs from %s sea-level (%04X).\r\n", png_name_0, sea_level_0, png_name_1, sea_level_1 );
		goto lbl_end;
	}


	// get prj to update
	if( GetTERR( prj_name_in, &info ) == FALSE ) goto lbl_end;
	if( ( width != info.header->width ) || ( height != info.header->height ) )
	{
		Printf( "Warning: %s heightmap dimensions (%dx%d) differ from new (%dx%d).\r\n", prj_name_in, info.header->width, info.header->height, width, height );
	}

	// prep
	layer1_size = ( ( ( ( ( width + 7 ) >> 3 ) * ( ( height + 7 ) >> 3 ) ) * sizeof( LAYER1::BLOCK ) ) * 2 );
	max_layer2_size = ( ( ( ( ( width + 7 ) >> 3 ) * ( ( height + 7 ) >> 3 ) ) * sizeof( LAYER2::BLOCK ) ) * 2 );
	max_out_buf_size = info.file_size - info.chunk_size + sizeof( TERR_HEADER ) + layer1_size + max_layer2_size + 8;

	out_buf = (BYTE*) VirtualAlloc( NULL, max_out_buf_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE );
	if( out_buf == NULL  ){ 
		Printf( "Error: Out of memory while processing %s.\r\n", prj_name_out );
		goto lbl_end;
	}
	header = (TERR_HEADER*) &out_buf[ (((BYTE*)info.header) - ((BYTE*)info.file_buf)) ];
	layer1 = (LAYER1*)(((BYTE*)header) + sizeof( TERR_HEADER ) );
	layer2 = (LAYER2*)(((BYTE*)layer1->blocks) + layer1_size );

	// fill out_buf //
	memcpy( out_buf, info.file_buf, ((BYTE*)header) - out_buf );
	memcpy( header, "TERR", 4 );
	header->width = width;
	header->height = height;
	header->num_blocks_layer1 = ( ( ( width + 7 ) >> 3 ) * ( ( height + 7 ) >> 3 ) );
	layer1->size = layer1_size;
	layer2->size = 0;
	
	// encode 
	block_index = 0;
	image = image_0;
	sea_level = sea_level_0;
	for(;;){ // run for both images
		for( DWORD y = 0; y < height; y+=8 ){
			for( DWORD x = 0; x < width; x+=8 ){
				// find base_height (min)
				WORD min = 0xFFFF;
				WORD max = 0x0000;
				for( DWORD yy = y; yy < min( y + 8, height ); yy++ ){
					for( DWORD xx = x; xx < min( x + 8, width ); xx++ ){
						WORD w = _byteswap_ushort( image[yy * width + xx] );
						min = min( min, w );
						max = max( max, w );
					}
				}
				if( max - min > 0x7F80 ){
					Printf( "Warning: Truncating elevation change for block starting at %d,%d\r\n", x,y );
				}
				layer1->blocks[block_index].base_height = ( ((DWORD)min) - ((DWORD)sea_level) );

				// build layer2 block
				BYTE layer2_block[8][8];
				for( DWORD yy = y; yy < min( y + 8, height ); yy++ ){
					for( DWORD xx = x; xx < min( x + 8, width ); xx++ ){
						WORD w = _byteswap_ushort( image[yy * width + xx] );
						layer2_block[ yy & 7 ][ xx & 7] = min( ((( w - min ) + 64 ) >> 7 ), 0x00FF );
					}
				}

				// search for duplicate
				DWORD off;
				for( off = 0; off < layer2->size; off += sizeof( LAYER2::BLOCK ) ) {
					if( memcmp( ((BYTE*)layer2->blocks) + off, layer2_block, sizeof( LAYER2::BLOCK ) ) == 0 ) break;
				}
				if( off == layer2->size ) // if dup not found
				{	
					memcpy( ((BYTE*)layer2->blocks) + layer2->size, layer2_block, sizeof( LAYER2::BLOCK ) );
					layer2->size += sizeof( LAYER2::BLOCK );
				}
				layer1->blocks[block_index].layer2_offset = off;

				block_index++;
			}
		}
		if( image == image_0 ){
			image = image_1;
			sea_level = sea_level_1;
		}
		else break;
	}

	header->num_blocks_layer2 = layer2->size / sizeof( LAYER2::BLOCK );
	header->size = sizeof(TERR_HEADER) + layer2->size + ( layer1->size >> 1 );
	memcpy( ((BYTE*)layer2->blocks) + layer2->size, ((BYTE*)info.layer2->blocks) + info.layer2->size, 
		( info.file_size - ( (DWORD)( ((BYTE*)info.header) - ((BYTE*)info.file_buf) ) + info.chunk_size ) )
	);
	out_size = ( info.file_size - info.chunk_size ) + ( sizeof(TERR_HEADER) + layer2->size + layer1->size + 8 );

	// write 
	bResult = SaveFile( prj_name_out, out_buf, out_size );
	if( bResult == FALSE ){
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", prj_name_out );
		PrintSystemError( dwErr );
	}

	// free
	VirtualFree( out_buf, 0, MEM_RELEASE );
lbl_end:
	if( info.file_buf != NULL ) VirtualFree( info.file_buf, 0, MEM_RELEASE );
	if( image_0 != NULL ) lodepng_free( image_0 );
	if( image_1 != NULL ) lodepng_free( image_1 );
	return bResult;
}



void SetSeaLevel( char* png_name, WORD sea_level )
{
	BYTE* png_buf;
	BYTE* out_buf;
	BYTE* chunk;
	DWORD png_size;
	DWORD len;

	png_buf = (BYTE*) ReadInFile( png_name, &png_size );
	if( png_buf == NULL ){
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", png_name );
		PrintSystemError( dwErr );
		return;
	}
	
	out_buf = (BYTE*) VirtualAlloc( NULL, png_size + 12, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE );
	if( out_buf == NULL )
	{
		Printf( "Error: Out of memory while processing %s.\r\n", png_name );
		VirtualFree( png_buf, 0, MEM_RELEASE );
		return;
	}

	memcpy( out_buf, png_buf, png_size );


	chunk = get_chunk( out_buf, png_size, 0x52526574 ); // "teRR"
	if( chunk == NULL ){ // create chunk
		chunk = get_chunk( png_buf, png_size, 0x444E4549 ); // "IEND"
		if( chunk == NULL ){ 
			Printf( "%s > Error: Bad PNG image file\r\n", png_name );
		}
		else {
			if( ( out_buf - 12 ) != chunk )
			{
				Printf( "%s > Error: Data after IEND chunk?\r\n", png_name );
			}
			else {
				png_size += 14;
				*((DWORD*)( &chunk[ 0 ]  )) = 0x02000000;
				*((DWORD*)( &chunk[ 4 ]  )) = 0x52526574; // "teRR" 
				*((WORD* )( &chunk[ 8 ]  )) = sea_level;
				*((DWORD*)( &chunk[ 10 ] )) = _byteswap_ulong( crc32( &chunk[ 4 ], 6 ) );
				*((DWORD*)( &chunk[ 14 ]  )) = 0x00000000; // IEND_len
				*((DWORD*)( &chunk[ 18 ] )) = 0x444E4549; // "IEND"
				*((DWORD*)( &chunk[ 22 ] )) = 0x826042AE; // IEND_crc
			}
		}
	}
	else {
		len = _byteswap_ulong( *((DWORD*)chunk) );
		*((WORD*)(&chunk[8])) = sea_level;
		*((DWORD*)(&chunk[ len + 8 ])) = _byteswap_ulong( crc32( &chunk[4], len + 4 ) );
	}
	
	if( SaveFile( png_name, out_buf, png_size ) == FALSE ) {
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", png_name );
		PrintSystemError( dwErr );
	}

	VirtualFree( out_buf, 0, MEM_RELEASE );
	VirtualFree( png_buf, 0, MEM_RELEASE );
}

void ShowSeaLevel( char* png_name )
{
	BYTE* png_buf;
	BYTE* chunk;
	DWORD png_size;

	// read-in file
	png_buf = (BYTE*) ReadInFile( png_name, &png_size );
	if( png_buf == NULL ){
		DWORD dwErr = GetLastError();
		Printf( "%s > Error: ", png_name );
		PrintSystemError( dwErr );
		return;
	}

	chunk = get_chunk( png_buf, png_size, 0x52526574 ); // "teRR"
	Printf( "%04X\r\n", ( ( chunk == NULL ) ? 0 : *((WORD*)(&chunk[8])) ) ); 

	VirtualFree( png_buf, 0, MEM_RELEASE );
}